import copy
import json
import time

from onshape_client.client import Client
from onshape_client.compatible_imports import parse, parse_qs
from onshape_client.oas import *
from onshape_client.units import u


class OnshapeElement(object):
    """ Turn a standard Onshape URL into an OnshapeElement object. Ensure that the URL is correctly formatted, and
    create the useful fields. Example of valid urls:
    https://cad.onshape.com/documents/c8f8013d34183b1de74fa930/w/574b77701d8b74987c273500/e/455ef770951fe37de0b8ff08
    https://cad.onshape.com/documents/c8f8013d34183b1de74fa930/w/574b77701d8b74987c273500/e/455ef770951fe37de0b8ff08?configuration=List_TOpkWtvolR0KY4%3Dewf
    """

    DRAWING_DATA_TYPE = "onshape-app/drawing"

    @staticmethod
    def create_from_oas_models(oas_model, **kwargs):
        if isinstance(oas_model, BTDocumentElementInfo):
            return OnshapeElement.create_from_ids(eid=oas_model.id, **kwargs)
        if isinstance(oas_model, BTDocumentInfo):
            return OnshapeElement.create_from_ids(did=oas_model.id, **kwargs)

    @staticmethod
    def create_from_ids(
        did=None,
        wvm=None,
        wvmid=None,
        eid=None,
        partid=None,
        configuration=None,
        sibling=None,
    ):
        client = Client.get_client()
        did = did if did else sibling.did
        wvm = wvm if wvm else sibling.wvm
        wvmid = wvmid if wvmid else sibling.wvmid

        url = client.configuration.host + "/documents/" + did + "/" + wvm + "/" + wvmid
        url = url + "/e/" + eid if eid else url
        url = url + "?configuration=" + configuration if configuration else url
        result = OnshapeElement(url)
        result.partid = partid
        return result

    @staticmethod
    def create(name="New Document"):
        """Returns a blank new document."""
        client = Client.get_client()
        doc_params = BTDocumentParams(name=name, is_public=True)
        doc = client.documents_api.create_document(doc_params)
        doc = OnshapeElement.create_from_ids(
            did=doc.id, wvm="w", wvmid=doc.default_workspace.id
        )
        return doc

    @staticmethod
    def create_from_file(file_path):
        client = Client.get_client()
        client.translation_api

    def import_file(self, file_path, **kwargs):
        """Import a file from the local file system. Returns the URL of the resulting element if translated."""
        client = Client.get_client()
        result = client.blob_elements_api.upload_file_create_element(
            self.did,
            self.wvmid,
            file=open(file_path, "rb"),
            translate=True,
            encoded_filename=file_path.name,
            **kwargs,
        )
        translation_id = result.translation_id
        result = OnshapeElement.poll_translation_result(translation_id)

        element_id = result.result_element_ids[0]
        return OnshapeElement.create_from_ids(eid=element_id, sibling=self)

    @staticmethod
    def poll_translation_result(translation_id):
        def is_polling_done(response):
            if response.request_state == "DONE":
                return True
            elif response.request_state == "ACTIVE":
                return False
            raise UserWarning(f"Translation failed")

        polling_function = lambda: Client.get_client().translation_api.get_translation(
            translation_id
        )
        return OnshapeElement.poll(polling_function, is_polling_done)

    def export_file(self, file_path, **kwargs):
        """Exports the element this class is pointing to"""
        if self._get_element_info().data_type == OnshapeElement.DRAWING_DATA_TYPE:
            result = Client.get_client().drawings_api.create_drawing_translation(
                did=self.did,
                wv=self.wvm,
                wvid=self.wvmid,
                eid=self.eid,
                bt_translate_format_params=BTTranslateFormatParams(
                    element_id=self.eid,
                    destination_name="exported_drawing",
                    format_name="PDF",
                    store_in_document=False,
                ),
            )
            translation_id = result.id
            result = OnshapeElement.poll_translation_result(translation_id)
            download_id = result.result_external_data_ids[0]
            file_path.write_bytes(
                Client.get_client()
                .documents_api.download_external_data(
                    did=self.did, fid=download_id, _preload_content=False
                )
                .data
            )

    def create_drawing(self, drawing_name="My Drawing"):
        """Create a four view drawing of the current element"""
        if self.wvm != "w":
            raise UserWarning(
                "Can only create a drawing in a workspace - not a version."
            )
        drawing = Client.get_client().drawings_api.create_drawing_app_element(
            bt_drawing_params=BTDrawingParams(
                document_id=self.did,
                workspace_id=self.wvmid,
                drawing_name=drawing_name,
                element_id=self.eid,
                views="four",
                template_document_id="cbe6e776694549b5ba1a3e88",
                template_workspace_id="24d08acf10234dbc8d3ab585",
                template_element_id="17eef7862b224f6fb12cbc46",
                projection="third",
                hidden_lines="EXCLUDED",
                is_surface=False,
                element_configuration="default",
                is_flattened_part=False,
                reference_type=1,
            )
        )
        drawing = OnshapeElement.create_from_ids(eid=drawing.id, sibling=self)
        return drawing

    @staticmethod
    def poll(polling_function, is_done_function):
        """For long polling for long-running jobs like translation. polling_function is the function that gets called
        every time step until is_done_function returns true."""

        result = polling_function()
        while not is_done_function(result):
            time.sleep(2)
            result = polling_function()
        return result

    def __init__(self, url, *args, **kwargs):
        self.original_url = url
        parsed_vals = parse(url)
        self.base_url = parsed_vals.scheme + "://" + parsed_vals.netloc
        path_list = parsed_vals.path.split("/")
        self.did = path_list[2]
        self.wvmid = path_list[4]
        self.wvm = path_list[3]
        length = len(path_list)
        if length > 8:
            eid = path_list[8]
            optional_microversion = path_list[6]
        elif length > 5 and path_list[5] == "e":
            eid = path_list[6]
            # This is the microversion retrieved from get_microversion_path() and represents a part that is pointing towards
            # both a version/workspace and a microversion.
            optional_microversion = None
        else:
            eid = None
            optional_microversion = None
        self.eid = eid
        if parsed_vals.query and "configuration" in parsed_vals.query:
            qs = parse_qs(parsed_vals.query)
            if len(qs["configuration"]) > 1:
                raise UserWarning(
                    "There should only be one configuration query in the url. Ignoring all but first."
                )
            self.configuration = qs["configuration"][0]
        else:
            self.configuration = None
        self.optional_microversion = optional_microversion

    def get_microversion_url(self):
        """Determine the microversion from the current version/workspace and return the path to that microversion. This
        will call the API to get the current microversion if the microversion is not already specified."""
        if self.optional_microversion:
            return self.get_url()
        else:
            res = Client.get_client().documents_api.get_current_microversion(
                self.did, self.wvm, self.wvmid, _preload_content=False
            )
            microversion = json.loads(res.data.decode("UTF-8"))["microversion"]
            self.optional_microversion = microversion
            return self.get_url()

    def get_url(self):
        url = self.base_url + "/documents/"
        if self.did:
            url = url + self.did + "/"
        if self.wvm:
            url = url + self.wvm + "/"
        if self.wvmid:
            url = url + self.wvmid + "/"
        if self.optional_microversion:
            url = url + "m/" + self.optional_microversion + "/"
        if self.eid:
            url = url + "e/" + self.eid + "/"
        if self.configuration:
            url = url + "?configuration=" + self.configuration + "/"
        return url[:-1]

    @property
    def element_type(self):
        elements = Client.get_client().documents_api.get_elements_in_document(
            self.did, self.wvm, self.wvmid
        )
        for element in elements:
            if element.id == self.eid:
                return element.type

    def elements(self, filter_name=None, filter_type=None, filter_data_type=None):
        result = []
        for e in self._get_element_infos():
            ands = []
            if filter_name and not filter_name == e.name:
                ands.append(False)
            if filter_type and not filter_type == e.element_type:
                ands.append(False)
            if filter_data_type and not filter_data_type == e.data_type:
                ands.append(False)
            if all(ands):
                result.append(
                    OnshapeElement.create_from_ids(
                        did=self.did, wvm=self.wvm, wvmid=self.wvmid, eid=e.id
                    )
                )
        return result

    @property
    def microversion(self):
        res = Client.get_client().documents_api.get_current_microversion(
            self.did, self.wvm, self.wvmid, _preload_content=False
        )
        microversion = json.loads(res.data.decode("UTF-8"))["microversion"]
        return microversion

    @property
    def name(self):
        if self.eid:
            return self._get_element_info().name
        return self._get_document_info().name

    @property
    def default_workspace(self):
        return self._get_document_info().default_workspace.id

    @property
    def assemblies(self):
        return self.elements(filter_type="ASSEMBLY")

    @property
    def part_studios(self):
        return self.elements(filter_type="PARTSTUDIO")

    @property
    def drawings(self):
        return self.elements(
            filter_type="APPLICATION", filter_data_type=OnshapeElement.DRAWING_DATA_TYPE
        )

    def s_assembly_insert_message(self):
        e_type = self.element_type
        message = BTAssemblyInstanceDefinitionParams(
            document_id=self.did,
            element_id=self.eid,
            is_assembly=e_type == "Assembly",
            is_whole_part_studio=e_type == "Part Studio",
        )
        if self.wvm == "w":
            message.microversion_id = self.microversion
        elif self.wvm == "v":
            message.version_id = self.wvmid
        elif self.wvm == "m":
            message.microversion_id = self.wvmid
        return message

    def new_assembly(self, name="Assembly"):
        asm = Client.get_client().assemblies_api.create_assembly(
            self.did,
            self.wvmid,
            BTModelElementParams(name=name),
            _preload_content=False,
        )
        asm = json.loads(asm.data.decode("utf-8"))
        return OnshapeElement.create_from_ids(
            did=self.did, wvm="w", wvmid=self.wvmid, eid=asm["id"]
        )

    def delete(self):
        Client.get_client().documents_api.delete_document(self.did)

    ########## Below methods call API methods and are not cached.

    def _get_element_info(self):
        return next(i for i in self._get_element_infos() if i.id == self.eid)

    def _get_document_info(self):
        return Client.get_client().documents_api.get_document(self.did)

    def _get_element_infos(self):
        return Client.get_client().documents_api.get_elements_in_document(
            self.did, self.wvm, self.wvmid
        )


class ConfiguredOnshapeElement(OnshapeElement):
    def __init__(self, *args, **kwargs):
        super(ConfiguredOnshapeElement, self).__init__(*args, **kwargs)
        # Initialize the cached values
        # Raw configuration params
        self._raw_configuration_params = self._get_raw_configuration_params()
        # Map from visible name to parameterId
        self._parameter_map = self._get_parameter_map()
        # Map from parameterId to default value
        self.default_configuration_map = self._get_default_configuration_map()
        # In the beginning, the current configuration is the default config
        self.current_configuration = self.default_configuration_map

    def get_url_with_configuration(self):
        """A configuration is applied on top of the current element.

        To be used like:
        >>> url = "https://cad.onshape.com/documents/cca81d10f239db0db9481e6f/v/ca51b7554314d6aab254d2e6/e/69c9eedda86512966b20bc90"
        >>> my_element = ConfiguredOnshapeElement(url)
        >>> my_element.update_current_configuration({"size": 20*u.m})
        >>> my_element.get_url_with_configuration()
        https://cad.onshape.com/documents/cca81d10f239db0db9481e6f/v/ca51b7554314d6aab254d2e6/e/69c9eedda86512966b20bc90?configuration=List_UKkGODiz574chc%3DDefault%3Bsize%3D20.0%2Bmeter
        >>> my_element.update_current_configuration({"Configuration": "chamfered"})
        >>> my_element.get_url_with_configuration()
        https://cad.onshape.com/documents/cca81d10f239db0db9481e6f/v/ca51b7554314d6aab254d2e6/e/69c9eedda86512966b20bc90?configuration=List_UKkGODiz574chc%3Dchamfered%3Bsize%3D20.0%2Bmeter
        """
        url = self.get_url() + "?" + self.get_configuration_query_param()
        return url

    def update_current_configuration(self, config):
        self.current_configuration.update(config)

    def get_configuration_string(self):
        encoded_val = self._get_configuration_encoding_response()
        return encoded_val["encodedId"]

    def get_configuration_query_param(self):
        encoded_val = self._get_configuration_encoding_response()
        return encoded_val["queryParam"]

    def _get_configuration_encoding_response(self):
        res = Client.get_client().elements_api.encode_configuration_map(
            self.did,
            self.eid,
            self._get_bt_configuration_params_for_current_configuration(),
            _preload_content=False,
        )
        return json.loads(res.data.decode("utf-8"))

    def _make_configuration_map(self, update):
        return self._get_default_configuration_map().update(update)

    def _get_default_configuration_map(self):
        defaults_map = {}
        for k, v in self._get_parameter_map().items():
            param_type = v["typeName"]
            if (
                param_type == "BTMConfigurationParameterEnum"
                or param_type == "BTMConfigurationParameterBoolean"
            ):
                default_value = v["message"]["defaultValue"]
            elif param_type == "BTMConfigurationParameterQuantity":
                range = v["message"]["rangeAndDefault"]["message"]
                default_value = u(str(range["defaultValue"]) + range["units"])
            else:
                raise NotImplementedError(
                    "This configuration type: {} is not supported.".format(param_type)
                )
            defaults_map[k] = default_value
        # Return a copy so the original doesn't get changed and we can keep using it
        self.default_configuration_map = defaults_map
        return copy.deepcopy(defaults_map)

    def _get_bt_configuration_params_for_current_configuration(self):
        final = []
        for k, v in self.current_configuration.items():
            p = ConfigurationEntry(
                parameter_id=self._get_param_id_from_name(k),
                parameter_value=str(v).replace(" ", "+"),
            )
            final.append(p)
        return BTConfigurationParams(parameters=final)

    def _get_param_id_from_name(self, name):
        try:
            return self._get_parameter_map()[name]["message"]["parameterId"]
        except KeyError as e:
            raise KeyError(
                "{} is not a valid parameter Id within the requested model configuration.".format(
                    e
                )
            )

    def _get_parameter_map(self):
        """Returns a map from the user-defined parameter names to the parameterIds ex: {"myEnumParam": "Listisds32d"}
        :return:
        """
        config_params = self._get_raw_configuration_params()
        parameter_map = {}
        for i, p in enumerate(config_params["configurationParameters"]):
            name = config_params["configurationParameters"][i]["message"][
                "parameterName"
            ]
            parameter_map[name] = config_params["configurationParameters"][i]
        return parameter_map

    def _get_raw_configuration_params(self):
        response = Client.get_client().elements_api.get_configuration(
            self.did, self.wvm, self.wvmid, self.eid, _preload_content=False
        )
        return json.loads(response.data.decode("utf-8"))



from __future__ import absolute_import

# import apis into api package
from onshape_client.oas.api.accounts_api import AccountsApi
from onshape_client.oas.api.app_elements_api import AppElementsApi
from onshape_client.oas.api.applications_api import ApplicationsApi
from onshape_client.oas.api.assemblies_api import AssembliesApi
from onshape_client.oas.api.billing_api import BillingApi
from onshape_client.oas.api.blob_elements_api import BlobElementsApi
from onshape_client.oas.api.companies_api import CompaniesApi
from onshape_client.oas.api.default_api import DefaultApi
from onshape_client.oas.api.documents_api import DocumentsApi
from onshape_client.oas.api.drawings_api import DrawingsApi
from onshape_client.oas.api.elements_api import ElementsApi
from onshape_client.oas.api.feature_studios_api import FeatureStudiosApi
from onshape_client.oas.api.folders_api import FoldersApi
from onshape_client.oas.api.metadata_api import MetadataApi
from onshape_client.oas.api.open_api_api import OpenAPIApi
from onshape_client.oas.api.part_studios_api import PartStudiosApi
from onshape_client.oas.api.parts_api import PartsApi
from onshape_client.oas.api.release_management_api import ReleaseManagementApi
from onshape_client.oas.api.revisions_api import RevisionsApi
from onshape_client.oas.api.teams_api import TeamsApi
from onshape_client.oas.api.translations_api import TranslationsApi
from onshape_client.oas.api.users_api import UsersApi
from onshape_client.oas.api.versions_api import VersionsApi
from onshape_client.oas.api.webhooks_api import WebhooksApi
from onshape_client.oas.api.workflow_api import WorkflowApi

# flake8: noqa



# coding: utf-8

# flake8: noqa
"""
    Onshape REST API

    The Onshape REST API consumed by all clients.  # noqa: E501

    The version of the OpenAPI document: 1.111
    Contact: api-support@onshape.zendesk.com
    Generated by: https://openapi-generator.tech
"""

# we can not import model classes here because that would create a circular
# reference which would not work in python2



def test_import_then_drawing_workflow(client, new_document, assets, tmp_dir):
    imported_element = new_document.import_file(file_path=assets / "Cube.x_t")
    drawing = imported_element.create_drawing()
    drawing.export_file(tmp_dir / "test_import_then_drawing_workflow.pdf")



# coding: utf-8

"""
    Onshape REST API

    The Onshape REST API consumed by all clients.  # noqa: E501

    The version of the OpenAPI document: 1.111
    Contact: api-support@onshape.zendesk.com
    Generated by: https://openapi-generator.tech
"""

from __future__ import absolute_import

import sys  # noqa: F401

import six  # noqa: F401
from onshape_client.oas.model_utils import (  # noqa: F401
    ModelNormal,
    datetime,
    int,
    str,
)

try:
    from onshape_client.oas.models import bt_billing_plan_info
except ImportError:
    bt_billing_plan_info = sys.modules["onshape_client.oas.models.bt_billing_plan_info"]
try:
    from onshape_client.oas.models import bt_card_info
except ImportError:
    bt_card_info = sys.modules["onshape_client.oas.models.bt_card_info"]
try:
    from onshape_client.oas.models import bt_plan_subscriber_info
except ImportError:
    bt_plan_subscriber_info = sys.modules[
        "onshape_client.oas.models.bt_plan_subscriber_info"
    ]
try:
    from onshape_client.oas.models import btapi_application_summary_info
except ImportError:
    btapi_application_summary_info = sys.modules[
        "onshape_client.oas.models.btapi_application_summary_info"
    ]
try:
    from onshape_client.oas.models import next_charge
except ImportError:
    next_charge = sys.modules["onshape_client.oas.models.next_charge"]
try:
    from onshape_client.oas.models import prorated_charges
except ImportError:
    prorated_charges = sys.modules["onshape_client.oas.models.prorated_charges"]
try:
    from onshape_client.oas.models import subscription
except ImportError:
    subscription = sys.modules["onshape_client.oas.models.subscription"]


class BTPurchaseInfo(ModelNormal):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Attributes:
      allowed_values (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          with a capitalized key describing the allowed value and an allowed
          value. These dicts store the allowed enum values.
      attribute_map (dict): The key is attribute name
          and the value is json key in definition.
      discriminator_value_class_map (dict): A dict to go from the discriminator
          variable value to the discriminator class name.
      validations (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          that stores validations for max_length, min_length, max_items,
          min_items, exclusive_maximum, inclusive_maximum, exclusive_minimum,
          inclusive_minimum, and regex.
      additional_properties_type (tuple): A tuple of classes accepted
          as additional properties values.
    """

    allowed_values = {}

    validations = {}

    additional_properties_type = None

    @staticmethod
    def openapi_types():
        """
        This must be a class method so a model may have properties that are
        of type self, this ensures that we don't create a cyclic import

        Returns
            openapi_types (dict): The key is attribute name
                and the value is attribute type.
        """
        return {
            "account_id": (str,),  # noqa: E501
            "actual_amount_paid_cents": (int,),  # noqa: E501
            "amount_cents": (int,),  # noqa: E501
            "application": (
                btapi_application_summary_info.BTAPIApplicationSummaryInfo,
            ),  # noqa: E501
            "canceled_at": (datetime,),  # noqa: E501
            "card": (bt_card_info.BTCardInfo,),  # noqa: E501
            "client_id": (str,),  # noqa: E501
            "coupon_amount_off": (int,),  # noqa: E501
            "coupon_percent_off": (int,),  # noqa: E501
            "created_by": (str,),  # noqa: E501
            "currency": (str,),  # noqa: E501
            "duration": (int,),  # noqa: E501
            "duration_months": (int,),  # noqa: E501
            "group": (str,),  # noqa: E501
            "href": (str,),  # noqa: E501
            "id": (str,),  # noqa: E501
            "last_modified": (datetime,),  # noqa: E501
            "last_modified_by": (str,),  # noqa: E501
            "light_seats": (int,),  # noqa: E501
            "name": (str,),  # noqa: E501
            "next_charge": (next_charge.NextCharge,),  # noqa: E501
            "payment_type": (int,),  # noqa: E501
            "pending_cancelation": (bool,),  # noqa: E501
            "plan": (bt_billing_plan_info.BTBillingPlanInfo,),  # noqa: E501
            "plan_id": (str,),  # noqa: E501
            "plan_name": (str,),  # noqa: E501
            "plan_type": (int,),  # noqa: E501
            "prorated_charges": ([prorated_charges.ProratedCharges],),  # noqa: E501
            "prorated_total": (int,),  # noqa: E501
            "purchase_date": (datetime,),  # noqa: E501
            "reseller_name": (str,),  # noqa: E501
            "seats": (int,),  # noqa: E501
            "state": (int,),  # noqa: E501
            "subscribers": (
                [bt_plan_subscriber_info.BTPlanSubscriberInfo],
            ),  # noqa: E501
            "subscription_begin_at": (datetime,),  # noqa: E501
            "subscription_end_at": (datetime,),  # noqa: E501
            "subscription_fields": (subscription.Subscription,),  # noqa: E501
            "subscription_id": (str,),  # noqa: E501
            "subscription_type": (int,),  # noqa: E501
            "trial_end": (datetime,),  # noqa: E501
            "trial_initiated_by": (str,),  # noqa: E501
            "view_ref": (str,),  # noqa: E501
        }

    @staticmethod
    def discriminator():
        return None

    attribute_map = {
        "account_id": "accountId",  # noqa: E501
        "actual_amount_paid_cents": "actualAmountPaidCents",  # noqa: E501
        "amount_cents": "amountCents",  # noqa: E501
        "application": "application",  # noqa: E501
        "canceled_at": "canceledAt",  # noqa: E501
        "card": "card",  # noqa: E501
        "client_id": "clientId",  # noqa: E501
        "coupon_amount_off": "couponAmountOff",  # noqa: E501
        "coupon_percent_off": "couponPercentOff",  # noqa: E501
        "created_by": "createdBy",  # noqa: E501
        "currency": "currency",  # noqa: E501
        "duration": "duration",  # noqa: E501
        "duration_months": "durationMonths",  # noqa: E501
        "group": "group",  # noqa: E501
        "href": "href",  # noqa: E501
        "id": "id",  # noqa: E501
        "last_modified": "lastModified",  # noqa: E501
        "last_modified_by": "lastModifiedBy",  # noqa: E501
        "light_seats": "lightSeats",  # noqa: E501
        "name": "name",  # noqa: E501
        "next_charge": "nextCharge",  # noqa: E501
        "payment_type": "paymentType",  # noqa: E501
        "pending_cancelation": "pendingCancelation",  # noqa: E501
        "plan": "plan",  # noqa: E501
        "plan_id": "planId",  # noqa: E501
        "plan_name": "planName",  # noqa: E501
        "plan_type": "planType",  # noqa: E501
        "prorated_charges": "proratedCharges",  # noqa: E501
        "prorated_total": "proratedTotal",  # noqa: E501
        "purchase_date": "purchaseDate",  # noqa: E501
        "reseller_name": "resellerName",  # noqa: E501
        "seats": "seats",  # noqa: E501
        "state": "state",  # noqa: E501
        "subscribers": "subscribers",  # noqa: E501
        "subscription_begin_at": "subscriptionBeginAt",  # noqa: E501
        "subscription_end_at": "subscriptionEndAt",  # noqa: E501
        "subscription_fields": "subscriptionFields",  # noqa: E501
        "subscription_id": "subscriptionId",  # noqa: E501
        "subscription_type": "subscriptionType",  # noqa: E501
        "trial_end": "trialEnd",  # noqa: E501
        "trial_initiated_by": "trialInitiatedBy",  # noqa: E501
        "view_ref": "viewRef",  # noqa: E501
    }

    @staticmethod
    def _composed_schemas():
        return None

    required_properties = set(
        [
            "_data_store",
            "_check_type",
            "_from_server",
            "_path_to_item",
            "_configuration",
        ]
    )

    def __init__(
        self,
        _check_type=True,
        _from_server=False,
        _path_to_item=(),
        _configuration=None,
        **kwargs
    ):  # noqa: E501
        """bt_purchase_info.BTPurchaseInfo - a model defined in OpenAPI


        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _from_server (bool): True if the data is from the server
                                False if the data is from the client (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            account_id (str): [optional]  # noqa: E501
            actual_amount_paid_cents (int): [optional]  # noqa: E501
            amount_cents (int): [optional]  # noqa: E501
            application (btapi_application_summary_info.BTAPIApplicationSummaryInfo): [optional]  # noqa: E501
            canceled_at (datetime): [optional]  # noqa: E501
            card (bt_card_info.BTCardInfo): [optional]  # noqa: E501
            client_id (str): [optional]  # noqa: E501
            coupon_amount_off (int): [optional]  # noqa: E501
            coupon_percent_off (int): [optional]  # noqa: E501
            created_by (str): [optional]  # noqa: E501
            currency (str): [optional]  # noqa: E501
            duration (int): [optional]  # noqa: E501
            duration_months (int): [optional]  # noqa: E501
            group (str): [optional]  # noqa: E501
            href (str): [optional]  # noqa: E501
            id (str): [optional]  # noqa: E501
            last_modified (datetime): [optional]  # noqa: E501
            last_modified_by (str): [optional]  # noqa: E501
            light_seats (int): [optional]  # noqa: E501
            name (str): [optional]  # noqa: E501
            next_charge (next_charge.NextCharge): [optional]  # noqa: E501
            payment_type (int): [optional]  # noqa: E501
            pending_cancelation (bool): [optional]  # noqa: E501
            plan (bt_billing_plan_info.BTBillingPlanInfo): [optional]  # noqa: E501
            plan_id (str): [optional]  # noqa: E501
            plan_name (str): [optional]  # noqa: E501
            plan_type (int): [optional]  # noqa: E501
            prorated_charges ([prorated_charges.ProratedCharges]): [optional]  # noqa: E501
            prorated_total (int): [optional]  # noqa: E501
            purchase_date (datetime): [optional]  # noqa: E501
            reseller_name (str): [optional]  # noqa: E501
            seats (int): [optional]  # noqa: E501
            state (int): [optional]  # noqa: E501
            subscribers ([bt_plan_subscriber_info.BTPlanSubscriberInfo]): [optional]  # noqa: E501
            subscription_begin_at (datetime): [optional]  # noqa: E501
            subscription_end_at (datetime): [optional]  # noqa: E501
            subscription_fields (subscription.Subscription): [optional]  # noqa: E501
            subscription_id (str): [optional]  # noqa: E501
            subscription_type (int): [optional]  # noqa: E501
            trial_end (datetime): [optional]  # noqa: E501
            trial_initiated_by (str): [optional]  # noqa: E501
            view_ref (str): [optional]  # noqa: E501
        """

        self._data_store = {}
        self._check_type = _check_type
        self._from_server = _from_server
        self._path_to_item = _path_to_item
        self._configuration = _configuration

        for var_name, var_value in six.iteritems(kwargs):
            setattr(self, var_name, var_value)



import json
from datetime import datetime

from onshape_client.client import Client
from onshape_client.utility import load_json


# For updating many properties of many elements
class BulkMetadataUpdate(object):
    def __init__(self, primary_metadata_body):
        # The primary metadata_body defines the wvm against which to update the metadata
        self.primary_metadata_body = primary_metadata_body
        self.additional_metadata_calls = []

    def add_metadata_body(self, metadata_body):
        self.additional_metadata_calls.append(metadata_body)

    def get_merged_bodies(self):
        body = self.primary_metadata_body.body
        # merge the bodies
        for metadata_body in self.additional_metadata_calls:
            body["items"].extend(metadata_body.body["items"])
        return body

    def send(self):
        body = self.get_merged_bodies()
        onshape_element = self.primary_metadata_body.onshape_element
        self.metadata = Client.get_client().metadata_api.update_wv_metadata(
            onshape_element.did,
            onshape_element.wvm,
            onshape_element.wvmid,
            body,
            _preload_content=False,
        )


# For updating multiple properties of a single Onshape element
class MetaDataBody(object):
    def __init__(self, onshape_element):
        self.body = {"items": []}
        self.onshape_element = onshape_element
        self.__element_property_map = None
        self.__part_property_map = None
        self.__part_index_to_part_id_map = None

    @property
    def element_property_map(self):
        if not self.__element_property_map:
            self.__element_property_map = (
                self.get_element_property_name_to_property_id_map()
            )
        return self.__element_property_map

    @property
    def part_index_to_pid_map(self):
        if not self.__part_index_to_part_id_map:
            (
                self.__part_property_map,
                self.__part_index_to_part_id_map,
            ) = self.get_part_property_name_to_property_id_map()
        return self.__part_index_to_part_id_map

    @property
    def part_property_map(self):
        if not self.__part_property_map:
            (
                self.__part_property_map,
                self.__part_index_to_part_id_map,
            ) = self.get_part_property_name_to_property_id_map()
        return self.__part_property_map

    def add_to_element_metadata(self, property_name, new_val, eid=None):
        """If needed, can specify other element metadata"""
        try:
            if not eid:
                eid = self.onshape_element.eid
            to_be_appended = {
                "properties": [
                    {
                        "value": new_val,
                        "propertyId": self.element_property_map[eid][property_name],
                    }
                ],
                "href": self.make_element_href(eid),
            }
            self.body["items"].append(to_be_appended)
        except:
            pass

    def add_to_part_metadata(self, property_name, new_val, pid=None, p_index=0):
        """Select part either by pid OR by the part index."""
        try:
            if not pid:
                pid = self.part_index_to_pid_map[p_index]
            to_be_appended = {
                "properties": [
                    {
                        "value": new_val,
                        "propertyId": self.part_property_map[pid][property_name],
                    }
                ],
                "href": self.make_part_href(pid),
            }
            self.body["items"].append(to_be_appended)
        except:
            pass

    def get_element_property_name_to_property_id_map(self):
        """Get the element metadata for the workspace/version and build the map"""
        client = Client.get_client()
        property_map = {}
        response = client.metadata_api.get_wv_es_metadata(
            self.onshape_element.did,
            self.onshape_element.wvm,
            self.onshape_element.wvmid,
            _preload_content=False,
        )
        data = json.loads(response.data)
        for element in data["items"]:
            eid = element["elementId"]
            property_map[eid] = {}
            props_to_be_added = property_map[eid]
            for property in element["properties"]:
                props_to_be_added[property["name"]] = property["propertyId"]
        return property_map

    def get_part_property_name_to_property_id_map(self):
        """Get the part property map for the element in the particular version and build out the property map"""
        client = Client.get_client()
        property_map = {}
        part_index_to_part_id = []
        el = self.onshape_element
        response = client.metadata_api.get_wmve_ps_metadata(
            el.did, el.wvm, el.wvmid, el.eid, _preload_content=False
        )
        data = load_json(response)
        for i, part in enumerate(data["items"]):
            partId = part["partId"]
            part_index_to_part_id.append(partId)
            property_map[partId] = {}
            props_to_be_added = property_map[partId]
            for property in part["properties"]:
                props_to_be_added[property["name"]] = property["propertyId"]
        return property_map, part_index_to_part_id

    def make_element_href(self, eid):
        return (
            Client.get_client().configuration.host
            + "/api/metadata/d/"
            + self.onshape_element.did
            + "/"
            + self.onshape_element.wvm
            + "/"
            + self.onshape_element.wvmid
            + "/e/"
            + eid
        )

    def make_part_href(self, pid):
        return self.make_element_href(self.onshape_element.eid) + "/p/{}".format(pid)

    def send(self):
        """Update the element metadata as seen fit."""
        self.metadata = Client.get_client().metadata_api.update_wv_metadata(
            self.onshape_element.did,
            self.onshape_element.wvm,
            self.onshape_element.wvmid,
            self.body,
        )


def update_one_element_metadata_example():
    from onshape_client.onshape_url import OnshapeElement

    print("Now updating a single property within a single element")
    element = OnshapeElement(
        "https://cad.onshape.com/documents/31e9fe5e53feb43633f360f4/w/052a282a976da47ba4c305dd/e/a270fe6f3d3ddc0dd6aee29c"
    )
    metadata_class = MetaDataBody(element)
    metadata_class.add_to_part_metadata(
        "Part number", "my glorious part number", pid="JHD"
    )
    metadata_class.send()


def update_many_element_metadata_example():
    from onshape_client.onshape_url import OnshapeElement

    print("Now updating multiple properties in many elements")

    time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    element1 = OnshapeElement(
        "https://cad.onshape.com/documents/31e9fe5e53feb43633f360f4/w/052a282a976da47ba4c305dd/e/a270fe6f3d3ddc0dd6aee29c"
    )
    metadata_class1 = MetaDataBody(element1)
    metadata_class1.add_to_part_metadata(
        "Part number", "my glorious part number! " + time, pid="JHD"
    )

    element2 = OnshapeElement(
        "https://cad.onshape.com/documents/31e9fe5e53feb43633f360f4/w/052a282a976da47ba4c305dd/e/d329ef11b727b097da599eb6"
    )
    metadata_class2 = MetaDataBody(element2)
    metadata_class2.add_to_part_metadata(
        "Description", "my wicked desciption! " + time, pid="JHD"
    )

    element3 = OnshapeElement(
        "https://cad.onshape.com/documents/31e9fe5e53feb43633f360f4/w/052a282a976da47ba4c305dd/e/e252cef9464938dc28356295"
    )
    metadata_class3 = MetaDataBody(element3)
    metadata_class3.add_to_part_metadata("Name", "my amazing name! " + time, pid="JHD")

    bulk_metadata = BulkMetadataUpdate(metadata_class1)
    bulk_metadata.add_metadata_body(metadata_class2)
    bulk_metadata.add_metadata_body(metadata_class3)

    bulk_metadata.send()


def test_set_metadata():
    update_one_element_metadata_example()
    update_many_element_metadata_example()



import json
import time
from datetime import datetime

import pytest
from onshape_client.assembly import AssemblyDefinition
from onshape_client.client import Client
from onshape_client.oas.models.bt_assembly_instance_definition_params import (
    BTAssemblyInstanceDefinitionParams,
)
from onshape_client.oas.models.bt_assembly_transformed_instances_definition_params import (
    BTAssemblyTransformedInstancesDefinitionParams,
)
from onshape_client.oas.models.bt_version_or_workspace_params import (
    BTVersionOrWorkspaceParams,
)
from onshape_client.oas.models.transform_group import TransformGroup
from onshape_client.onshape_url import OnshapeElement
from onshape_client.utility import get_field


@pytest.mark.parametrize("element", ["asm_three_axes"], indirect=True)
def test_assembly_definition(client, element):
    """Simple api call without serialization."""
    res = client.assemblies_api.get_assembly_definition(
        element.did,
        "v",
        element.wvmid,
        element.eid,
        include_mate_features=True,
        include_mate_connectors=True,
        _preload_content=False,
    )
    parts = get_field(res, "parts")
    assert parts[0]["mateConnectors"]


@pytest.mark.parametrize("element", ["asm_three_axes"], indirect=True)
def test_roundtrip_instance_insert(client, element, new_document):
    """Attempt to create an assembly from a 'prototype' by using the assembly definition to insert the various
    parts/assemblies/definitions. This is also useful for performance testing. Currently only works for flat assemblies.
    :param client: client to make the API calls.
    :param element: an assembly to duplicate into a new document.
    """
    definition = client.assemblies_api.get_assembly_definition(
        element.did,
        element.wvm,
        element.wvmid,
        element.eid,
        include_mate_features=True,
        include_mate_connectors=True,
        _preload_content=False,
    )
    definition = json.loads(definition.data.decode("UTF-8"))
    from_assembly_definition = AssemblyDefinition(definition, origin_element=element)
    to_assembly = new_document.assemblies[0]
    client.assemblies_api.insert_transformed_instances(
        to_assembly.did,
        to_assembly.eid,
        to_assembly.wvmid,
        BTAssemblyTransformedInstancesDefinitionParams(
            transform_groups=from_assembly_definition.get_as_transform_groups()
        ),
        _preload_content=False,
    )
    new_definition = client.assemblies_api.get_assembly_definition(
        to_assembly.did,
        to_assembly.wvm,
        to_assembly.wvmid,
        to_assembly.eid,
        _preload_content=False,
    )
    new_definition = AssemblyDefinition(json.loads(new_definition.data))

    assert from_assembly_definition.is_structurally_equal_assembly(new_definition)


@pytest.mark.parametrize("element", ["ps_configurable_cube"], indirect=True)
def test_create_instance(client, element, new_copied_document_factory):
    doc = new_copied_document_factory(element)
    assembly = doc.new_assembly()
    insertable = doc.elements(filter_name="Cylinder")[0]
    response = client.assemblies_api.create_instance(
        assembly.did,
        assembly.wvmid,
        assembly.eid,
        bt_assembly_instance_definition_params=insertable.s_assembly_insert_message(),
        _preload_content=False,
    )
    assert response.status == 200


@pytest.mark.parametrize("element", ["asm_three_axes"], indirect=True)
@pytest.mark.skip("Not a highly critical test - and currently broken.")
def test_sub_sub_sub_assembly_instance_insert(client, element):
    """Ensures inner assemblies are inserted as expected by inserting sub-assembly at +1 m in X, then inserting that
    assembly at +1 in X, and sees if we ended up back at the origin."""

    def make_version(name, did):
        return json.loads(
            Client.get_client()
            .documents_api.create_version(
                did,
                bt_version_or_workspace_params=BTVersionOrWorkspaceParams(
                    name=name + "-" + str(datetime.now()), document_id=did
                ),
                _preload_content=False,
            )
            .data
        )["id"]

    origin = OnshapeElement(
        "https://cad.onshape.com/documents/ba13458e7ebfd1755645d5d1/w/031ba2bcb4b13b77bc5c43f7/e/af7fd0b2eac7ad63477370be"
    )
    assembly_pushed_x = OnshapeElement(
        "https://cad.onshape.com/documents/ba13458e7ebfd1755645d5d1/w/031ba2bcb4b13b77bc5c43f7/e/b340f69741da2eaba07ddc1f"
    )
    assembly_pulled_x = OnshapeElement(
        "https://cad.onshape.com/documents/ba13458e7ebfd1755645d5d1/w/031ba2bcb4b13b77bc5c43f7/e/6c9bb5a752c0b5b66610b918"
    )

    origin_transform = [
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        1.0,
    ]

    push_x_transform = [
        1.0,
        0.0,
        0.0,
        1.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        1.0,
    ]

    pull_x_transform = [
        1.0,
        0.0,
        0.0,
        -1.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        1.0,
    ]

    three_axes_instance = BTAssemblyInstanceDefinitionParams(
        document_id=element.did,
        version_id=element.wvmid,
        element_id=element.eid,
        is_assembly=True,
    )

    client.assemblies_api.insert_transformed_instances(
        origin.did,
        origin.eid,
        origin.wvmid,
        BTAssemblyTransformedInstancesDefinitionParams(
            transform_groups=[
                TransformGroup(
                    transform=origin_transform, instances=[three_axes_instance]
                )
            ]
        ),
    )

    origin_version = make_version("orogin", origin.did)

    origin_instance = BTAssemblyInstanceDefinitionParams(
        document_id=origin.did,
        version_id=origin_version,
        element_id=origin.eid,
        is_assembly=True,
    )

    client.assemblies_api.insert_transformed_instances(
        assembly_pushed_x.did,
        assembly_pushed_x.eid,
        assembly_pushed_x.wvmid,
        BTAssemblyTransformedInstancesDefinitionParams(
            transform_groups=[
                TransformGroup(transform=push_x_transform, instances=[origin_instance])
            ]
        ),
    )

    pushed_x_version = make_version("pushed_x", origin.did)

    assembly_pushed_x_instance = BTAssemblyInstanceDefinitionParams(
        document_id=assembly_pushed_x.did,
        version_id=pushed_x_version,
        element_id=assembly_pushed_x.eid,
        is_assembly=True,
    )

    client.assemblies_api.insert_transformed_instances(
        assembly_pulled_x.did,
        assembly_pulled_x.eid,
        assembly_pulled_x.wvmid,
        BTAssemblyTransformedInstancesDefinitionParams(
            transform_groups=[
                TransformGroup(
                    transform=pull_x_transform, instances=[assembly_pushed_x_instance]
                )
            ]
        ),
    )

    result = client.assemblies_api.get_assembly_definition(
        assembly_pulled_x.did,
        assembly_pulled_x.wvm,
        assembly_pulled_x.wvmid,
        assembly_pulled_x.eid,
        _preload_content=False,
    )
    result = json.loads(result.data.decode("UTF-8"))
    final_transform = [
        t["transform"]
        for t in result["rootAssembly"]["occurrences"]
        if len(t["path"]) == 4
    ][0]
    assert final_transform == origin_transform


def test_assembly_definition_performance(client):
    n_calls = 10
    calls = []
    cube_asm = OnshapeElement(
        "https://cad.onshape.com/documents/ee5f3dea1b2b6196cba97e4a/v/6935ab731cfe24915c2b4ac8/e/2e54df6ea6c0657369a57131"
    )
    for i in range(n_calls):
        start_time = time.time()
        client.assemblies_api.get_assembly_definition(
            cube_asm.did,
            cube_asm.wvm,
            cube_asm.wvmid,
            cube_asm.eid,
            _preload_content=False,
        )
        calls.append(time.time() - start_time)
    assert sum(calls) / len(calls) < 1.5


@pytest.mark.skip("Not yet implemented")
def test_stl_performance(client):
    cube_asm = OnshapeElement(
        "https://cad.onshape.com/documents/ee5f3dea1b2b6196cba97e4a/v/6935ab731cfe24915c2b4ac8/e/23e18c935aaf18a2dcba486b"
    )
    assert (
        performance_test_calls(
            lambda: client.part_studios_api.get_(
                cube_asm.did,
                cube_asm.wvm,
                cube_asm.wvmid,
                cube_asm.eid,
                partid="JHD",
                _preload_content=False,
            )
        )
        < 1
    )


def test_metadata_performance(client):
    cube_asm = OnshapeElement(
        "https://cad.onshape.com/documents/ee5f3dea1b2b6196cba97e4a/v/6935ab731cfe24915c2b4ac8/e/23e18c935aaf18a2dcba486b"
    )
    assert (
        performance_test_calls(
            lambda: client.parts_api.get_part_metadata(
                cube_asm.did,
                cube_asm.wvm,
                cube_asm.wvmid,
                cube_asm.eid,
                partid="JHD",
                _preload_content=False,
            )
        )
        < 1
    )


def performance_test_calls(call):
    n_calls = 10
    calls = []
    for i in range(n_calls):
        start_time = time.time()
        call()
        calls.append(time.time() - start_time)
    average = sum(calls) / len(calls)
    print(average)
    return average



"""API Modeling example that shows how to generate a pipe programmatically."""

import json

from onshape_client.client import Client
from onshape_client.oas import (
    BTMParameterQueryList148,
    BTMIndividualQuery138,
    BTCurveGeometryCircle115,
    BTMSketchCurve4,
    BTMSketch151,
    BTFeatureDefinitionCall1406,
    BTDocumentParams,
    BTFeatureScriptEvalCall2377,
    BTMParameterEnum145,
    BTMParameterQuantity147,
    BTMFeature134,
)

# Setup the client and pass the keys. Note that secret keys can be specified directly, OR they can be
# specified in environment variables OR in a yaml file at the 'keys_file' location. Here we use the file at the default
# 'keys_file' location.
from onshape_client.onshape_url import OnshapeElement

client = Client()

# create a new document
MY_NEW_DOCUMENT = "My new circle document"
doc_params = BTDocumentParams(name=MY_NEW_DOCUMENT)
doc = client.documents_api.create_document(doc_params)
doc = OnshapeElement.create_from_ids(
    did=doc.id, wvm="w", wvmid=doc.default_workspace.id
)

# Get the default Part Studio in the document
part_studio = None
elements = client.documents_api.get_elements_in_document(doc.did, doc.wvm, doc.wvmid)
for element in elements:
    if element.type == "Part Studio":
        element = OnshapeElement.create_from_oas_models(
            element, did=doc.did, wvmid=doc.wvmid, wvm=doc.wvm
        )
        part_studio = element
        break
if not part_studio:
    raise AttributeError("This document is unexpectedly missing a part studio.")
print(f"Your new part studio is: {part_studio.get_url()}")

# Make the circle object
FEATURE_NAME = "My New Circle"
PLANE_ID = "JDC"  # The plane deterministic ID for the sketch... This is known to be the id for the Top plane.
plane_query = BTMParameterQueryList148(
    parameter_id="sketchPlane",
    queries=[BTMIndividualQuery138(deterministic_ids=[PLANE_ID])],
)
CIRCLE_ID = "myCircle"

circle_geometry = BTCurveGeometryCircle115(
    radius=0.5, xcenter=0.0, ycenter=0.0, xdir=1.0, ydir=0.0, clockwise=False
)
circle = BTMSketchCurve4(
    center_id=f"{CIRCLE_ID}.center", entity_id=CIRCLE_ID, geometry=circle_geometry
)
sketch = BTMSketch151(entities=[circle], name=FEATURE_NAME, parameters=[plane_query])
feature_definition = BTFeatureDefinitionCall1406(feature=sketch)

# Pass the circle object
result_feature = client.part_studios_api.add_part_studio_feature(
    did=part_studio.did,
    wvm=part_studio.wvm,
    wvmid=part_studio.wvmid,
    eid=part_studio.eid,
    bt_feature_definition_call1406_bt_feature_definition_call1406=feature_definition,
)

# Query for the deterministic ID of the created sketch region using an FS query. For more info on how to use different queries, look at: https://cad.onshape.com/FsDoc/library.html#module-query.fs and https://forum.onshape.com/discussions/tagged/featurescript-query/p1
get_sketch_face_fs = f"""
function(context, queries)
{{
    return transientQueriesToStrings(evaluateQuery(context, qCreatedBy(makeId("{result_feature.feature.feature_id}"), EntityType.FACE)))[0];
}}"""
deterministic_ids_response = client.part_studios_api.eval_feature_script(
    did=part_studio.did,
    wvm=part_studio.wvm,
    wvmid=part_studio.wvmid,
    eid=part_studio.eid,
    bt_feature_script_eval_call2377_bt_feature_script_eval_call2377=BTFeatureScriptEvalCall2377(
        script=get_sketch_face_fs
    ),
    _preload_content=False,
    _check_input_type=False,
)
det_id = json.loads(deterministic_ids_response.data)["result"]["message"]["value"]

# Generate an extrude feature
tool_body_type = BTMParameterEnum145(
    value="SOLID", enum_name="ToolBodyType", parameter_id="bodyType"
)
operation_type = BTMParameterEnum145(
    value="NEW", enum_name="NewBodyOperationType", parameter_id="operationType"
)
region_query = BTMParameterQueryList148(
    parameter_id="entities", queries=[BTMIndividualQuery138(deterministic_ids=[det_id])]
)
length = BTMParameterQuantity147(expression="3.03*in", parameter_id="depth")
extrude_feature = BTMFeature134(
    bt_type="BTMFeature-134",
    name="My extrude",
    feature_type="extrude",
    parameters=[tool_body_type, operation_type, region_query, length],
)
feature_definition = BTFeatureDefinitionCall1406(feature=extrude_feature)
result_feature_extrude = client.part_studios_api.add_part_studio_feature(
    did=part_studio.did,
    wvm=part_studio.wvm,
    wvmid=part_studio.wvmid,
    eid=part_studio.eid,
    bt_feature_definition_call1406_bt_feature_definition_call1406=feature_definition,
)



"""Get all the FeatureScripts within a given document and save them to a local folder.
"""

import os.path
from pathlib import Path

from onshape_client.onshape_url import OnshapeElement


def test_get_all_fs(client):
    # We're saving these files to the home directory ("~")
    client.configuration.temp_folder_path = os.path.expanduser("~")

    # Turn the URL into an "OnshapeElement"
    url = "https://cad.onshape.com/documents/ea1c0fef91272adff0cc4b41/w/65efb40b6df7b9206b2a10b5/e/6921aecc8424c978fe8c78ce"
    cube = OnshapeElement(url, client=client)

    my_elements = client.documents_api.get_elements3(cube.did, cube.wvm, cube.wvmid)
    for element in my_elements:
        if element.element_type == "FEATURESTUDIO":
            fs_contents = client.feature_studios_api.get_feature_studio_contents1(
                cube.did, cube.wvm, cube.wvmid, element.id
            )
            file_path = (
                Path(__file__).parent
                / "tmp"
                / ("{name}::{id}.fs".format(name=element.name, id=element.id))
            )
            file = file_path.open(mode="w")
            file.write(fs_contents.contents)
            file.close()



# coding: utf-8

"""
    Onshape REST API

    The Onshape REST API consumed by all clients.  # noqa: E501

    The version of the OpenAPI document: 1.111
    Contact: api-support@onshape.zendesk.com
    Generated by: https://openapi-generator.tech
"""

from __future__ import absolute_import

# python 2 and python 3 compatibility library
import six
from onshape_client.oas.api_client import ApiClient
from onshape_client.oas.exceptions import ApiTypeError, ApiValueError
from onshape_client.oas.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    file_type,
    int,
    none_type,
    str,
    validate_and_convert_types,
)
from onshape_client.oas.models import bt_user_o_auth2_summary_info
from onshape_client.oas.models import bt_user_settings_info


class UsersApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __get_user_settings(self, uid, **kwargs):
            """get_user_settings  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True
            >>> thread = api.get_user_settings(uid, async_req=True)
            >>> result = thread.get()

            Args:
                uid (str):

            Keyword Args:
                includematerials (bool): [optional] if omitted the server will use the default value of True
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int): specifies the index of the server
                    that we want to use.
                    Default is 0.
                async_req (bool): execute request asynchronously

            Returns:
                bt_user_settings_info.BTUserSettingsInfo
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs["async_req"] = kwargs.get("async_req", False)
            kwargs["_return_http_data_only"] = kwargs.get(
                "_return_http_data_only", True
            )
            kwargs["_preload_content"] = kwargs.get("_preload_content", True)
            kwargs["_request_timeout"] = kwargs.get("_request_timeout", None)
            kwargs["_check_input_type"] = kwargs.get("_check_input_type", True)
            kwargs["_check_return_type"] = kwargs.get("_check_return_type", True)
            kwargs["_host_index"] = kwargs.get("_host_index", 0)
            kwargs["uid"] = uid
            return self.call_with_http_info(**kwargs)

        self.get_user_settings = Endpoint(
            settings={
                "response_type": (bt_user_settings_info.BTUserSettingsInfo,),
                "auth": [],
                "endpoint_path": "/api/users/{uid}/settings",
                "operation_id": "get_user_settings",
                "http_method": "GET",
                "servers": [],
            },
            params_map={
                "all": ["uid", "includematerials",],
                "required": ["uid",],
                "nullable": [],
                "enum": [],
                "validation": [],
            },
            root_map={
                "validations": {},
                "allowed_values": {},
                "openapi_types": {"uid": (str,), "includematerials": (bool,),},
                "attribute_map": {
                    "uid": "uid",
                    "includematerials": "includematerials",
                },
                "location_map": {"uid": "path", "includematerials": "query",},
                "collection_format_map": {},
            },
            headers_map={
                "accept": [
                    "application/vnd.onshape.v1+json;charset=UTF-8;qs=0.1",
                    "application/json;charset=UTF-8; qs=0.09",
                ],
                "content_type": [],
            },
            api_client=api_client,
            callable=__get_user_settings,
        )

        def __get_user_settings_current_logged_in_user(self, **kwargs):
            """get_user_settings_current_logged_in_user  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True
            >>> thread = api.get_user_settings_current_logged_in_user(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                includematerials (bool): [optional] if omitted the server will use the default value of True
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int): specifies the index of the server
                    that we want to use.
                    Default is 0.
                async_req (bool): execute request asynchronously

            Returns:
                bt_user_settings_info.BTUserSettingsInfo
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs["async_req"] = kwargs.get("async_req", False)
            kwargs["_return_http_data_only"] = kwargs.get(
                "_return_http_data_only", True
            )
            kwargs["_preload_content"] = kwargs.get("_preload_content", True)
            kwargs["_request_timeout"] = kwargs.get("_request_timeout", None)
            kwargs["_check_input_type"] = kwargs.get("_check_input_type", True)
            kwargs["_check_return_type"] = kwargs.get("_check_return_type", True)
            kwargs["_host_index"] = kwargs.get("_host_index", 0)
            return self.call_with_http_info(**kwargs)

        self.get_user_settings_current_logged_in_user = Endpoint(
            settings={
                "response_type": (bt_user_settings_info.BTUserSettingsInfo,),
                "auth": [],
                "endpoint_path": "/api/users/settings",
                "operation_id": "get_user_settings_current_logged_in_user",
                "http_method": "GET",
                "servers": [],
            },
            params_map={
                "all": ["includematerials",],
                "required": [],
                "nullable": [],
                "enum": [],
                "validation": [],
            },
            root_map={
                "validations": {},
                "allowed_values": {},
                "openapi_types": {"includematerials": (bool,),},
                "attribute_map": {"includematerials": "includematerials",},
                "location_map": {"includematerials": "query",},
                "collection_format_map": {},
            },
            headers_map={
                "accept": [
                    "application/vnd.onshape.v1+json;charset=UTF-8;qs=0.1",
                    "application/json;charset=UTF-8; qs=0.09",
                ],
                "content_type": [],
            },
            api_client=api_client,
            callable=__get_user_settings_current_logged_in_user,
        )

        def __session_info(self, **kwargs):
            """session_info  # noqa: E501

            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True
            >>> thread = api.session_info(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int): specifies the index of the server
                    that we want to use.
                    Default is 0.
                async_req (bool): execute request asynchronously

            Returns:
                bt_user_o_auth2_summary_info.BTUserOAuth2SummaryInfo
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs["async_req"] = kwargs.get("async_req", False)
            kwargs["_return_http_data_only"] = kwargs.get(
                "_return_http_data_only", True
            )
            kwargs["_preload_content"] = kwargs.get("_preload_content", True)
            kwargs["_request_timeout"] = kwargs.get("_request_timeout", None)
            kwargs["_check_input_type"] = kwargs.get("_check_input_type", True)
            kwargs["_check_return_type"] = kwargs.get("_check_return_type", True)
            kwargs["_host_index"] = kwargs.get("_host_index", 0)
            return self.call_with_http_info(**kwargs)

        self.session_info = Endpoint(
            settings={
                "response_type": (
                    bt_user_o_auth2_summary_info.BTUserOAuth2SummaryInfo,
                ),
                "auth": [],
                "endpoint_path": "/api/users/sessioninfo",
                "operation_id": "session_info",
                "http_method": "GET",
                "servers": [],
            },
            params_map={
                "all": [],
                "required": [],
                "nullable": [],
                "enum": [],
                "validation": [],
            },
            root_map={
                "validations": {},
                "allowed_values": {},
                "openapi_types": {},
                "attribute_map": {},
                "location_map": {},
                "collection_format_map": {},
            },
            headers_map={
                "accept": [
                    "application/vnd.onshape.v1+json;charset=UTF-8;qs=0.1",
                    "application/json;charset=UTF-8; qs=0.09",
                ],
                "content_type": [],
            },
            api_client=api_client,
            callable=__session_info,
        )


class Endpoint(object):
    def __init__(
        self,
        settings=None,
        params_map=None,
        root_map=None,
        headers_map=None,
        api_client=None,
        callable=None,
    ):
        """Creates an endpoint

        Args:
            settings (dict): see below key value pairs
                'response_type' (tuple/None): response type
                'auth' (list): a list of auth type keys
                'endpoint_path' (str): the endpoint path
                'operation_id' (str): endpoint string identifier
                'http_method' (str): POST/PUT/PATCH/GET etc
                'servers' (list): list of str servers that this endpoint is at
            params_map (dict): see below key value pairs
                'all' (list): list of str endpoint parameter names
                'required' (list): list of required parameter names
                'nullable' (list): list of nullable parameter names
                'enum' (list): list of parameters with enum values
                'validation' (list): list of parameters with validations
            root_map
                'validations' (dict): the dict mapping endpoint parameter tuple
                    paths to their validation dictionaries
                'allowed_values' (dict): the dict mapping endpoint parameter
                    tuple paths to their allowed_values (enum) dictionaries
                'openapi_types' (dict): param_name to openapi type
                'attribute_map' (dict): param_name to camelCase name
                'location_map' (dict): param_name to  'body', 'file', 'form',
                    'header', 'path', 'query'
                collection_format_map (dict): param_name to `csv` etc.
            headers_map (dict): see below key value pairs
                'accept' (list): list of Accept header strings
                'content_type' (list): list of Content-Type header strings
            api_client (ApiClient) api client instance
            callable (function): the function which is invoked when the
                Endpoint is called
        """
        self.settings = settings
        self.params_map = params_map
        self.params_map["all"].extend(
            [
                "async_req",
                "_host_index",
                "_preload_content",
                "_request_timeout",
                "_return_http_data_only",
                "_check_input_type",
                "_check_return_type",
            ]
        )
        self.params_map["nullable"].extend(["_request_timeout"])
        self.validations = root_map["validations"]
        self.allowed_values = root_map["allowed_values"]
        self.openapi_types = root_map["openapi_types"]
        extra_types = {
            "async_req": (bool,),
            "_host_index": (int,),
            "_preload_content": (bool,),
            "_request_timeout": (none_type, int, (int,), [int]),
            "_return_http_data_only": (bool,),
            "_check_input_type": (bool,),
            "_check_return_type": (bool,),
        }
        self.openapi_types.update(extra_types)
        self.attribute_map = root_map["attribute_map"]
        self.location_map = root_map["location_map"]
        self.collection_format_map = root_map["collection_format_map"]
        self.headers_map = headers_map
        self.api_client = api_client
        self.callable = callable

    def __validate_inputs(self, kwargs):
        for param in self.params_map["enum"]:
            if param in kwargs:
                check_allowed_values(self.allowed_values, (param,), kwargs[param])

        for param in self.params_map["validation"]:
            if param in kwargs:
                check_validations(self.validations, (param,), kwargs[param])

        if kwargs["_check_input_type"] is False:
            return

        for key, value in six.iteritems(kwargs):
            fixed_val = validate_and_convert_types(
                value,
                self.openapi_types[key],
                [key],
                False,
                kwargs["_check_input_type"],
                configuration=self.api_client.configuration,
            )
            kwargs[key] = fixed_val

    def __gather_params(self, kwargs):
        params = {
            "body": None,
            "collection_format": {},
            "file": {},
            "form": [],
            "header": {},
            "path": {},
            "query": [],
        }

        for param_name, param_value in six.iteritems(kwargs):
            param_location = self.location_map.get(param_name)
            if param_location is None:
                continue
            if param_location:
                if param_location == "body":
                    params["body"] = param_value
                    continue
                base_name = self.attribute_map[param_name]
                if param_location == "form" and self.openapi_types[param_name] == (
                    file_type,
                ):
                    params["file"][param_name] = [param_value]
                elif param_location == "form" and self.openapi_types[param_name] == (
                    [file_type],
                ):
                    # param_value is already a list
                    params["file"][param_name] = param_value
                elif param_location in {"form", "query"}:
                    param_value_full = (base_name, param_value)
                    params[param_location].append(param_value_full)
                if param_location not in {"form", "query"}:
                    params[param_location][base_name] = param_value
                collection_format = self.collection_format_map.get(param_name)
                if collection_format:
                    params["collection_format"][base_name] = collection_format

        return params

    def __call__(self, *args, **kwargs):
        """ This method is invoked when endpoints are called
        Example:
        pet_api = PetApi()
        pet_api.add_pet  # this is an instance of the class Endpoint
        pet_api.add_pet()  # this invokes pet_api.add_pet.__call__()
        which then invokes the callable functions stored in that endpoint at
        pet_api.add_pet.callable or self.callable in this class
        """
        return self.callable(self, *args, **kwargs)

    def call_with_http_info(self, **kwargs):

        try:
            _host = self.settings["servers"][kwargs["_host_index"]]
        except IndexError:
            if self.settings["servers"]:
                raise ApiValueError(
                    "Invalid host index. Must be 0 <= index < %s"
                    % len(self.settings["servers"])
                )
            _host = None

        for key, value in six.iteritems(kwargs):
            if key not in self.params_map["all"]:
                raise ApiTypeError(
                    "Got an unexpected parameter '%s'"
                    " to method `%s`" % (key, self.settings["operation_id"])
                )
            # only throw this nullable ApiValueError if _check_input_type
            # is False, if _check_input_type==True we catch this case
            # in self.__validate_inputs
            if (
                key not in self.params_map["nullable"]
                and value is None
                and kwargs["_check_input_type"] is False
            ):
                raise ApiValueError(
                    "Value may not be None for non-nullable parameter `%s`"
                    " when calling `%s`" % (key, self.settings["operation_id"])
                )

        for key in self.params_map["required"]:
            if key not in kwargs.keys():
                raise ApiValueError(
                    "Missing the required parameter `%s` when calling "
                    "`%s`" % (key, self.settings["operation_id"])
                )

        self.__validate_inputs(kwargs)

        params = self.__gather_params(kwargs)

        accept_headers_list = self.headers_map["accept"]
        if accept_headers_list:
            params["header"]["Accept"] = self.api_client.select_header_accept(
                accept_headers_list
            )

        content_type_headers_list = self.headers_map["content_type"]
        if content_type_headers_list:
            header_list = self.api_client.select_header_content_type(
                content_type_headers_list
            )
            params["header"]["Content-Type"] = header_list

        return self.api_client.call_api(
            self.settings["endpoint_path"],
            self.settings["http_method"],
            params["path"],
            params["query"],
            params["header"],
            body=params["body"],
            post_params=params["form"],
            files=params["file"],
            response_type=self.settings["response_type"],
            auth_settings=self.settings["auth"],
            async_req=kwargs["async_req"],
            _check_type=kwargs["_check_return_type"],
            _return_http_data_only=kwargs["_return_http_data_only"],
            _preload_content=kwargs["_preload_content"],
            _request_timeout=kwargs["_request_timeout"],
            _host=_host,
            collection_formats=params["collection_format"],
        )



from onshape_client.client import Client


def test_specify_client_config_in_code():
    client = Client(
        configuration={
            "base_url": "https://cad.onshape.com",
            "access_key": "USER_ACCESS_KEY",
            "secret_key": "USER_SECRET_KEY",
        }
    )

    assert client.configuration.api_key["SECRET_KEY"] == "USER_SECRET_KEY"
    assert client.configuration.api_key["ACCESS_KEY"] == "USER_ACCESS_KEY"

    # Reset client:
    client = Client()
    assert client.configuration.api_key["SECRET_KEY"] != "USER_SECRET_KEY"
    assert client.configuration.api_key["ACCESS_KEY"] != "USER_ACCESS_KEY"



import pytest
from onshape_client.oas import BTPartMetadataInfo, BTPartAppearanceInfo, BTColorInfo
import webbrowser
import json


@pytest.mark.parametrize("element", ["ps_configurable_cube"], indirect=True)
def test_change_color(client, element, new_copied_document_factory):
    """Test to change part color."""
    new_cube = new_copied_document_factory(element).part_studios[0]
    meta_data = BTPartMetadataInfo(
        appearance=BTPartAppearanceInfo(color=BTColorInfo(blue=0, green=0, red=255))
    )
    parts = client.parts_api.get_parts_wmve(
        did=new_cube.did, wvm=new_cube.wvm, wvmid=new_cube.wvmid, eid=new_cube.eid
    )
    client.parts_api.update_part_metadata(
        did=new_cube.did,
        wvm=new_cube.wvm,
        wvmid=new_cube.wvmid,
        eid=new_cube.eid,
        partid=parts[0].part_id,
        body=json.dumps([meta_data.to_dict()]),
    )
    parts = client.parts_api.get_parts_wmve(
        did=new_cube.did, wvm=new_cube.wvm, wvmid=new_cube.wvmid, eid=new_cube.eid
    )
    assert parts[0].appearance.color.red == 255



"""Given a model and a desired configuration, generate the link to get to the configuration.
"""

from onshape_client.onshape_url import ConfiguredOnshapeElement
from onshape_client.units import u


def test_create_configuration_link():
    # A configurable cube URL
    url = "https://cad.onshape.com/documents/cca81d10f239db0db9481e6f/v/aaa25d18038e06f0b2964e2d/e/69c9eedda86512966b20bc90"
    element = ConfiguredOnshapeElement(url)

    # Desired configuration to apply to the cube
    new_config_map = {
        "size": 10 * u.inch,
        "edge_configuration": "chamfered",
        "show_cube": True,
    }
    element.update_current_configuration(new_config_map)
    print(
        "Link to model with the desired configuration applied: "
        + element.get_url_with_configuration()
    )



import pytest


@pytest.mark.parametrize("element", ["ps_configurable_cube"], indirect=True)
def test_export_drawing(client, element, tmp_dir):
    drawing = element.drawings[0]
    translate_result = drawing.export_file(tmp_dir / "test_export.pdf")



# coding: utf-8

"""
    Onshape REST API

    The Onshape REST API consumed by all clients.  # noqa: E501

    OpenAPI spec version: 1.96
    Contact: api-support@onshape.zendesk.com
    Generated by: https://openapi-generator.tech
"""

from setuptools import setup, find_packages  # noqa: H301
import os

NAME = "onshape_client"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = [
    "urllib3 >= 1.15",
    "six >= 1.10",
    "certifi",
    "python-dateutil",
    "requests_oauthlib",
    "ruamel.yaml",
    "pathlib",
    "pint",
]

setup(
    name=NAME,
    version=os.environ["ONSHAPE_CLIENT_VERSION"],
    description="Onshape REST API",
    author_email="api-support@onshape.zendesk.com",
    url="",
    keywords=["OpenAPI", "OpenAPI-Generator", "Onshape REST API"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description="""\
    The Onshape REST API consumed by all clients.  # noqa: E501
    """,
)



# coding: utf-8

"""
    Onshape REST API

    The Onshape REST API consumed by all clients.  # noqa: E501

    The version of the OpenAPI document: 1.111
    Contact: api-support@onshape.zendesk.com
    Generated by: https://openapi-generator.tech
"""

from __future__ import absolute_import

import six  # noqa: F401
from onshape_client.oas.model_utils import (  # noqa: F401
    ModelNormal,
    str,
)


class JsonNode(ModelNormal):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Attributes:
      allowed_values (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          with a capitalized key describing the allowed value and an allowed
          value. These dicts store the allowed enum values.
      attribute_map (dict): The key is attribute name
          and the value is json key in definition.
      discriminator_value_class_map (dict): A dict to go from the discriminator
          variable value to the discriminator class name.
      validations (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          that stores validations for max_length, min_length, max_items,
          min_items, exclusive_maximum, inclusive_maximum, exclusive_minimum,
          inclusive_minimum, and regex.
      additional_properties_type (tuple): A tuple of classes accepted
          as additional properties values.
    """

    allowed_values = {
        ("node_type",): {
            "ARRAY": "ARRAY",
            "BINARY": "BINARY",
            "BOOLEAN": "BOOLEAN",
            "MISSING": "MISSING",
            "NULL": "NULL",
            "NUMBER": "NUMBER",
            "OBJECT": "OBJECT",
            "POJO": "POJO",
            "STRING": "STRING",
        },
    }

    validations = {}

    additional_properties_type = None

    @staticmethod
    def openapi_types():
        """
        This must be a class method so a model may have properties that are
        of type self, this ensures that we don't create a cyclic import

        Returns
            openapi_types (dict): The key is attribute name
                and the value is attribute type.
        """
        return {
            "array": (bool,),  # noqa: E501
            "big_decimal": (bool,),  # noqa: E501
            "big_integer": (bool,),  # noqa: E501
            "binary": (bool,),  # noqa: E501
            "boolean": (bool,),  # noqa: E501
            "container_node": (bool,),  # noqa: E501
            "double": (bool,),  # noqa: E501
            "empty": (bool,),  # noqa: E501
            "float": (bool,),  # noqa: E501
            "floating_point_number": (bool,),  # noqa: E501
            "int": (bool,),  # noqa: E501
            "integral_number": (bool,),  # noqa: E501
            "long": (bool,),  # noqa: E501
            "missing_node": (bool,),  # noqa: E501
            "node_type": (str,),  # noqa: E501
            "null": (bool,),  # noqa: E501
            "number": (bool,),  # noqa: E501
            "object": (bool,),  # noqa: E501
            "pojo": (bool,),  # noqa: E501
            "short": (bool,),  # noqa: E501
            "textual": (bool,),  # noqa: E501
            "value_node": (bool,),  # noqa: E501
        }

    @staticmethod
    def discriminator():
        return None

    attribute_map = {
        "array": "array",  # noqa: E501
        "big_decimal": "bigDecimal",  # noqa: E501
        "big_integer": "bigInteger",  # noqa: E501
        "binary": "binary",  # noqa: E501
        "boolean": "boolean",  # noqa: E501
        "container_node": "containerNode",  # noqa: E501
        "double": "double",  # noqa: E501
        "empty": "empty",  # noqa: E501
        "float": "float",  # noqa: E501
        "floating_point_number": "floatingPointNumber",  # noqa: E501
        "int": "int",  # noqa: E501
        "integral_number": "integralNumber",  # noqa: E501
        "long": "long",  # noqa: E501
        "missing_node": "missingNode",  # noqa: E501
        "node_type": "nodeType",  # noqa: E501
        "null": "null",  # noqa: E501
        "number": "number",  # noqa: E501
        "object": "object",  # noqa: E501
        "pojo": "pojo",  # noqa: E501
        "short": "short",  # noqa: E501
        "textual": "textual",  # noqa: E501
        "value_node": "valueNode",  # noqa: E501
    }

    @staticmethod
    def _composed_schemas():
        return None

    required_properties = set(
        [
            "_data_store",
            "_check_type",
            "_from_server",
            "_path_to_item",
            "_configuration",
        ]
    )

    def __init__(
        self,
        _check_type=True,
        _from_server=False,
        _path_to_item=(),
        _configuration=None,
        **kwargs
    ):  # noqa: E501
        """json_node.JsonNode - a model defined in OpenAPI


        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _from_server (bool): True if the data is from the server
                                False if the data is from the client (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            array (bool): [optional]  # noqa: E501
            big_decimal (bool): [optional]  # noqa: E501
            big_integer (bool): [optional]  # noqa: E501
            binary (bool): [optional]  # noqa: E501
            boolean (bool): [optional]  # noqa: E501
            container_node (bool): [optional]  # noqa: E501
            double (bool): [optional]  # noqa: E501
            empty (bool): [optional]  # noqa: E501
            float (bool): [optional]  # noqa: E501
            floating_point_number (bool): [optional]  # noqa: E501
            int (bool): [optional]  # noqa: E501
            integral_number (bool): [optional]  # noqa: E501
            long (bool): [optional]  # noqa: E501
            missing_node (bool): [optional]  # noqa: E501
            node_type (str): [optional]  # noqa: E501
            null (bool): [optional]  # noqa: E501
            number (bool): [optional]  # noqa: E501
            object (bool): [optional]  # noqa: E501
            pojo (bool): [optional]  # noqa: E501
            short (bool): [optional]  # noqa: E501
            textual (bool): [optional]  # noqa: E501
            value_node (bool): [optional]  # noqa: E501
        """

        self._data_store = {}
        self._check_type = _check_type
        self._from_server = _from_server
        self._path_to_item = _path_to_item
        self._configuration = _configuration

        for var_name, var_value in six.iteritems(kwargs):
            setattr(self, var_name, var_value)



import os
import warnings
import webbrowser
from enum import Enum
from pathlib import Path
from types import SimpleNamespace

from oauthlib.oauth2 import (
    UnauthorizedClientError,
    UnsupportedGrantTypeError,
    MissingTokenError,
    InvalidClientIdError,
)
from onshape_client.oas import ApiClient
from onshape_client.oas import api
from onshape_client.oas.configuration import Configuration
from requests_oauthlib import OAuth2Session
from ruamel.yaml import YAML

from .oauth.local_server import start_server


def add_defaults(namespace, **kwargs):
    for config_name, env_var_name in kwargs.items():
        add_default(namespace, config_name, env_var_name)
    return namespace


def add_default(namespace, config_name, env_var_name):
    new_config_var = ConfigVar(config_name, env_var_name)
    if new_config_var.config_name not in vars(namespace):
        setattr(namespace, new_config_var.config_name, new_config_var)


# Only update if the config doesn't already have the value AND the default has the value not evaluated as False
def update_with_default_carefully(config, default_config):
    for k, v in default_config.items():
        if v and (k not in config or not config[k]):
            config[k] = v


class ConfigVar:
    def __init__(self, config_name, env_var_name):
        self.config_name = config_name
        self.env_var_name = env_var_name


class Client:
    """
    Initialize an Onshape client.

    The default is to pull the configuration from a file located in the user's home
    directory. However, configs can also be passed directly as a dictionary.

    Keys can be obtained from https://dev-portal.onshape.com.

    Parameters
    ----------
    conf_file : :obj:`str`, optional
        The path to the location of an onshape_client_config YAML file, where multiple stacks may be declared.
    configuration : :obj:`dict`, optional
        This overrides the conf_file parameter by supplying a dictionary that looks like the following:
        {
            "base_url": "https://cad.onshape.com",
            "access_key": "YOUR_ACCESS_KEY_HERE",
            "secret_key": "YOUR_SECRET_KEY_HERE"
        }
    stack_key : :obj:`str`, optional
        The key of the name for the particular stack to be used. Only applicable when reading configuration from the
        conf_file.
    """

    singleton_instance = None

    config_vars = SimpleNamespace()

    prod_base_url = "https://cad.onshape.com"

    # Mapping from configuration property to shell variable in order to allow passing from environment context.
    add_defaults(
        config_vars,
        access_key="ONSHAPE_API_ACCESS_KEY",
        authorization_uri="ONSHAPE_AUTHORIZATION_URI",
        base_url="ONSHAPE_BASE_URL",
        client_id="ONSHAPE_CLIENT_ID",
        client_secret="ONSHAPE_CLIENT_SECRET",
        oauth_authorization_method="ONSHAPE_AUTHORIZATION_METHOD",
        redirect_uri="ONSHAPE_REDIRECT_URI",
        scope="ONSHAPE_SCOPE",
        secret_key="ONSHAPE_API_SECRET_KEY",
        token_uri="ONSHAPE_TOKEN_URI",
    )

    @staticmethod
    def get_client():
        """ get an singleton_instance of the client class singleton. """
        client = Client.singleton_instance
        if not client:
            raise Exception("Please manually instantiate the client.")
        return client

    @staticmethod
    def clear_client():
        """ delete the singleton singleton_instance of the client. """
        Client.singleton_instance = None

    @staticmethod
    def _get_from_environment():
        defaults = {}
        for name, config_var in vars(Client.config_vars).items():
            if config_var.env_var_name in os.environ:
                defaults.setdefault(
                    config_var.config_name, os.environ[config_var.env_var_name]
                )
        return defaults

    @staticmethod
    def _get_configuration_from_keys_file(keys_file, stack_key):
        config = {}
        try:
            yaml = YAML()
            configurations_file = yaml.load(Path(keys_file))
            config = configurations_file[
                stack_key if stack_key else configurations_file["default_stack"]
            ]

        except KeyError as e:
            raise KeyError(
                "Your creds file is not constructed as expected. The key: {} was expected and not found.".format(
                    e
                )
            )
        except FileNotFoundError:
            pass
        return config

    def __init__(
        self,
        keys_file="~/.onshape_client_config.yaml",
        configuration=None,
        stack_key=None,
        open_authorize_grant_callback=None,
    ):
        """Configuration values can come from setting directly, with configuration= or from environment variables, or
        from the configuration file specified with key_file in that order without overwriting. Those in the key_file are
        considered defaults.

        :param keys_file:
        :param configuration:
        :param stack_key:
        :param open_authorize_grant_callback: This will get called when the application needs to show the user the grant
            authorization page. It gets passed the URL to go to. Once the grant is authorized, the application needs to
            call client.set_grant_authorization_url_response(redirected_url) with whatever url the user got directed to
            (that includes the authorization code!)
        """
        if Client.singleton_instance:
            warnings.warn(
                "A Client was already created so this will create another and override it. Please use "
                "Client.get_client() to get the previously created client."
            )
        final_configuration = configuration if configuration else {}
        update_with_default_carefully(
            final_configuration, Client._get_from_environment()
        )
        if keys_file:
            keys_file = os.path.expanduser(keys_file)
            update_with_default_carefully(
                final_configuration,
                Client._get_configuration_from_keys_file(keys_file, stack_key),
            )
        self.stack_key = stack_key
        self._set_configuration(final_configuration)
        self._create_apis()
        self.open_authorize_grant_callback = open_authorize_grant_callback

        if self.get_authentication_method() == "oauth":
            self._set_oauth_session()

        Client.singleton_instance = self

    def set_grant_authorization_url_response(self, authorization_url_response):
        """ This is the redirected-to url, for example: https:localhost/oauth_redirect?code=XraRXPYGSWfZmBvUIiNHvSlZ&state=kgZQJbVp731CgNxAeKJN7TIoOlo5bz"""
        self.fetch_access_token(authorization_url_response)

    def get_authentication_method(self):
        # Prefer OAUTH if specified, otherwise try for apikeys
        if self.client_secret and self.client_id:
            return "oauth"
        elif (
            self.configuration.api_key["SECRET_KEY"]
            and self.configuration.api_key["ACCESS_KEY"]
        ):
            return "api_keys"
        else:
            return None

    def get_authorization_url(self):
        url, state = self.oauth.authorization_url(self.authorization_uri)
        return url

    def do_oauth_flow(self):
        """Do the oauth flow to set the access token"""
        try:
            self.refresh_access_token()
            # InvalidClientIdError : when the refresh token isn't valid
        except (
            UnauthorizedClientError,
            UnsupportedGrantTypeError,
            MissingTokenError,
            InvalidClientIdError,
        ) as e:
            authorization_method = OAuthAuthorizationMethods(
                self.oauth_authorization_method
            )
            oauth_type = OAuthAuthorizationMethods(self.oauth_authorization_method)
            if oauth_type == OAuthAuthorizationMethods.MANUAL_FLOW:
                raise OAuthNotAuthorizedException(
                    "You need to have the user authorize the grant."
                )
            elif authorization_method == OAuthAuthorizationMethods.LOCALHOST_SERVER:
                start_server(self.fetch_access_token, self.open_authorize_grant)
            elif authorization_method == OAuthAuthorizationMethods.PYTHON_CALLBACK:
                self.open_authorize_grant()

    def open_authorize_grant(self):
        done = self.fetch_access_token
        callback = self.open_authorize_grant_callback
        oauth_type = OAuthAuthorizationMethods(self.oauth_authorization_method)
        url = self.get_authorization_url()
        defaults_supported = [OAuthAuthorizationMethods.LOCALHOST_SERVER]
        if not callback and oauth_type not in defaults_supported:
            raise NotImplementedError(
                "To use OAuth, you need to pass in a callback to the client constructor with open_authorize_grant_callback="
            )
        elif callback:
            try:
                if callback and not callable(callback):
                    # Attempt to evaluate the function definition
                    callback = eval(callback)
                callback(url, done)
            except:
                raise AttributeError(
                    "The open_authorize_grant_callback function did not work."
                )
        else:
            self.default_callback(url)

    def default_callback(self, url):
        authorization_method = OAuthAuthorizationMethods(
            self.oauth_authorization_method
        )
        if authorization_method == OAuthAuthorizationMethods.LOCALHOST_SERVER:
            webbrowser.open(url)

    def _set_configuration(self, configuration_dictionary):
        configuration = Configuration()

        configuration.api_key["SECRET_KEY"] = configuration_dictionary.get(
            "secret_key", ""
        )
        configuration.api_key["ACCESS_KEY"] = configuration_dictionary.get(
            "access_key", ""
        )

        self.client_id = configuration_dictionary.get("client_id", None)
        self.client_secret = configuration_dictionary.get("client_secret", None)
        configuration.access_token = configuration_dictionary.get("access_token", "")
        self.refresh_token = configuration_dictionary.get("refresh_token", None)
        self.scope = configuration_dictionary.get("scope", None)
        self.redirect_uri = configuration_dictionary.get("redirect_uri", None)
        self.token_uri = configuration_dictionary.get(
            "token_uri", "https://oauth.onshape.com/oauth/token"
        )
        self.authorization_uri = configuration_dictionary.get(
            "authorization_uri", "https://oauth.onshape.com/oauth/authorize"
        )
        self.oauth_authorization_method = configuration_dictionary.get(
            "oauth_authorization_method", OAuthAuthorizationMethods.MANUAL_FLOW
        )
        # Default to prod
        configuration.host = configuration_dictionary.get(
            "base_url", "https://cad.onshape.com"
        )
        self.configuration = configuration
        return

    def _set_oauth_session(self):
        self.oauth = OAuth2Session(
            self.client_id, redirect_uri=self.redirect_uri, scope=self.scope
        )
        return

    """Override this method to get the resulting access code dictionary"""

    def refresh_access_token(self):
        # So that the oauth library doesn't complain about scope changing
        os.environ["OAUTHLIB_RELAX_TOKEN_SCOPE"] = "True"

        token_response = self.oauth.refresh_token(
            self.token_uri,
            refresh_token=self.refresh_token,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        self._set_oauth_creds_from_token_response(token_response)
        return token_response

    def fetch_access_token(self, **kwargs):
        """
        :param authorization_response: The response containing the authorization code.
        :return: The access token and the refresh token. These should be stored by the caller to facilitate future
        client initialization.
        """
        try:
            token_response = self.oauth.fetch_token(
                self.token_uri, client_secret=self.client_secret, **kwargs
            )
            self._set_oauth_creds_from_token_response(token_response)
            return token_response
        except Warning:
            # This is probably a scope warning.
            pass

    def _set_oauth_creds_from_token_response(self, token_response):
        self.configuration.access_token = token_response["access_token"]
        self.refresh_token = token_response["refresh_token"]
        return

    def _create_apis(self):
        api_client = ApiClient(configuration=self.configuration)
        self.api_client = api_client
        self.accounts_api = api.AccountsApi(api_client)
        self.app_elements_api = api.AppElementsApi(api_client)
        self.assemblies_api = api.AssembliesApi(api_client)
        self.blob_elements_api = api.BlobElementsApi(api_client)
        self.drawings_api = api.DrawingsApi(api_client)
        self.documents_api = api.DocumentsApi(api_client)
        self.elements_api = api.ElementsApi(api_client)
        self.feature_studios_api = api.FeatureStudiosApi(api_client)
        self.metadata_api = api.MetadataApi(api_client)
        self.parts_api = api.PartsApi(api_client)
        self.part_studios_api = api.PartStudiosApi(api_client)
        self.translation_api = api.TranslationsApi(api_client)


class OAuthAuthorizationMethods(Enum):
    # This is fully implemented, and meant for locally hosted applications (desktop apps)
    LOCALHOST_SERVER = "localhost_server"

    # This is for when the open_authorize_grant_callback is set. It yields control to this client
    # by specifying a function that should be called to authorize the grant, and passes in a
    # function to that callback that needs to be called when the grant has been authorized with
    # the result.
    PYTHON_CALLBACK = "python_callback"

    # This flow will communicate that the client needs the authorization flow with an OAuthNotAuthorizedException
    # after failing to refresh the token. Once the client throws the error, the application should
    # call client.get_authorization_url() and present the resulting url to the user, and subsequently
    # call client.fetch_access_token(resulting_url_with_authorization_code) to complete the flow.
    # Note that for the best user experience, the client should store the refresh_token in persistent
    # storage so the user only has to go through the grant authorization step once.
    MANUAL_FLOW = "manual_flow"


# Exception that should be caught signaling that the user should go through the grant authorization flow.
class OAuthNotAuthorizedException(Exception):
    def __init__(self, *args, **kwargs):
        super(OAuthNotAuthorizedException, self).__init__(args, kwargs)



def test_import_file(new_document, client, assets):
    files_to_import = ["Cube.x_t", "Cube.step"]
    for file in files_to_import:
        imported_ps = new_document.import_file(assets / file)
        response = client.parts_api.get_mass_properties(
            imported_ps.did,
            imported_ps.wvm,
            imported_ps.wvmid,
            imported_ps.eid,
            partid="JFD",
        )
        assert response.bodies["JFD"].volume[0] > 0
